/*
PFS is Copyright (C) 2001 Douglas Thain.
This program is released under the GNU General Public License.
See the file COPYING for details.
*/

#include "pfs_file.h"
#include "pfs_file_cache.h"
#include "pfs_service.h"
#include "pfs_resolve.h"

extern "C" {
#include "debug.h"
#include "stringtools.h"
#include "username.h"
#include "domain_name_cache.h"
#include "copy_stream.h"
#include "file_cache.h"
#include "full_io.h"
}

#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <errno.h>
#include <stdlib.h>
#include <utime.h>

extern struct file_cache *pfs_file_cache;
extern int pfs_sloppy_cache;

#define BUFFER_SIZE 65536

static int copy_fd_to_file( int fd, pfs_file *file )
{
	int ractual, wactual, offset = 0;
	char buffer[BUFFER_SIZE];

	while(1) {
		ractual = full_pread(fd,buffer,sizeof(buffer),offset);
		if(ractual<=0) return ractual;

		wactual = file->write(buffer,ractual,offset);
		if(wactual>=0) {
			offset += wactual;
		} else {
			return -1;
		}
	}
}

static int copy_file_to_fd( pfs_file *file, int fd )
{
	int ractual, wactual, offset = 0;
	char buffer[BUFFER_SIZE];

	while(1) {
		ractual = file->read(buffer,sizeof(buffer),offset);
		if(ractual<=0) return ractual;

		wactual = full_pwrite(fd,buffer,ractual,offset);
		if(wactual>=0) {
			offset += wactual;
		} else {
			return -1;
		}
	}
}

class pfs_file_cached : public pfs_file
{
private:
	int fd;
	int mode;
	int changed;

public:
	pfs_file_cached( pfs_name *n, int f, int m ) : pfs_file(n) {
		fd = f;
		mode = m;
		changed = 0;
	}

	virtual int close() {
		int result = -1;
		if(changed) {
			debug(D_CACHE,"storing %s",name.path);
			pfs_file *wfile = name.service->open(&name,O_WRONLY|O_CREAT|O_TRUNC,mode);
			if(wfile) {
				if(copy_fd_to_file(fd,wfile)==0) {
					result = 0;
				} else {
					result = -1;
				}
				wfile->close();
				delete wfile;
			} else {
				result = -1;
			}
		} else {
			result = 0;
		}
		/* give it a dummy truncate to update the mtime */
		/* this will prevent a later fetch of the same file */
		::ftruncate(fd,this->get_size());
		::close(fd);
		return result;
	}

	virtual pfs_ssize_t read( void *d, pfs_size_t length, pfs_off_t offset ) {
		return ::full_pread(fd,d,length,offset);
	}

	virtual pfs_ssize_t write( const void *d, pfs_size_t length, pfs_off_t offset ) {
		changed = 1;
		return ::full_pwrite(fd,d,length,offset);
	}

	virtual int fstat( struct stat *buf ) {
		return ::fstat(fd,buf);
	}

	virtual int fstatfs( struct statfs *buf ) {
		return ::fstatfs(fd,buf);
	}

	virtual int ftruncate( pfs_size_t length ) {
		changed = 1;
		return ::ftruncate(fd,length);
	}

	virtual pfs_ssize_t get_size() {
		struct stat buf;
		if(this->fstat(&buf)==0) {
			return buf.st_size;
		} else {
			return 0;
		}
	}

	virtual int get_local_name( char *n ) {
		return file_cache_contains(pfs_file_cache,name.path,n);
	}

	virtual int is_seekable() {
		return 1;
	}
};

pfs_file * pfs_cache_open( pfs_name *name, int flags, mode_t mode )
{
	struct stat buf;
	char txn[PFS_PATH_MAX];
	int fd, ok_to_fail;
	struct pfs_file *rfile, *result;
	struct utimbuf ut;

	if((name->flags & PFS_RESOLVE_FLAGS_SLOPPYCACHE) || pfs_sloppy_cache ) {
		buf.st_size = 0;
		buf.st_mtime = 0;
	} else {
		if(name->service->stat(name,&buf)!=0) {
			if(flags&O_CREAT && errno==ENOENT) {
				buf.st_mtime = 0;
				buf.st_size = 0;
			} else {
				return 0;
			}
		}
	}
	
	fd = file_cache_open(pfs_file_cache,name->path,txn,buf.st_size,buf.st_mtime);
	if(fd>=0) return new pfs_file_cached(name,fd,mode);

	debug(D_CACHE,"loading %s",name->path);

	fd = file_cache_begin(pfs_file_cache,name->path,txn);
	if(fd<0) return 0;

	if(flags&O_TRUNC) {
		rfile = 0;
		ok_to_fail = 1;
	} else if(flags&O_CREAT) {
		rfile = name->service->open(name,O_RDONLY,0);
		ok_to_fail = 1;
	} else {
		rfile = name->service->open(name,O_RDONLY,0);
		ok_to_fail = 0;
	}

	if(rfile) {
		if(copy_file_to_fd(rfile,fd)==0) {
			/* set the times to that of the remote file */
			ut.actime = buf.st_atime;
			ut.modtime = buf.st_mtime;
			::utime(txn,&ut);
			if(file_cache_commit(pfs_file_cache,name->path,txn)==0) {
				result = new pfs_file_cached(name,fd,mode);
			} else {
				result = 0;
			}
		} else {
			result = 0;
		}
		int save_errno = errno;
		rfile->close();
		delete rfile;
		errno = save_errno;
	} else if(ok_to_fail) {
		result = new pfs_file_cached(name,fd,mode);
		if(result) result->ftruncate(0);
	} else {
		result = 0;
	}

	if(result) {
		return result;
	} else {
		close(fd);
		file_cache_abort(pfs_file_cache,name->path,txn);
		return 0;
	}
}
